@@ -25,11 +25,13 @@
 #include <wlr/types/wlr_data_device.h>
 #include <wlr/types/wlr_export_dmabuf_v1.h>
 #include <wlr/types/wlr_gamma_control_v1.h>
-#include <wlr/types/wlr_idle.h>
 #include <wlr/types/wlr_idle_inhibit_v1.h>
+#include <wlr/types/wlr_idle_notify_v1.h>
 #include <wlr/types/wlr_output_layout.h>
 #include <wlr/types/wlr_output_management_v1.h>
 #include <wlr/types/wlr_presentation_time.h>
+#include <wlr/types/wlr_primary_selection_v1.h>
+#include <wlr/types/wlr_relative_pointer_v1.h>
 #include <wlr/types/wlr_scene.h>
 #include <wlr/types/wlr_screencopy_v1.h>
 #include <wlr/types/wlr_server_decoration.h>
@@ -59,10 +61,28 @@
 #include "xwayland.h"
 #endif
 
+void
+server_terminate(struct cg_server *server)
+{
+	// Workaround for https://gitlab.freedesktop.org/wayland/wayland/-/merge_requests/421
+	if (server->terminated) {
+		return;
+	}
+
+	wl_display_terminate(server->wl_display);
+}
+
+static void
+handle_display_destroy(struct wl_listener *listener, void *data)
+{
+	struct cg_server *server = wl_container_of(listener, server, display_destroy);
+	server->terminated = true;
+}
+
 static int
 sigchld_handler(int fd, uint32_t mask, void *data)
 {
-	struct wl_display *display = data;
+	struct cg_server *server = data;
 
 	/* Close Cage's read pipe. */
 	close(fd);
@@ -73,7 +93,8 @@ sigchld_handler(int fd, uint32_t mask, void *data)
 		wlr_log(WLR_DEBUG, "Connection closed by server");
 	}
 
-	wl_display_terminate(display);
+	server->return_app_code = true;
+	server_terminate(server);
 	return 0;
 }
 
@@ -97,7 +118,7 @@ set_cloexec(int fd)
 }
 
 static bool
-spawn_primary_client(struct wl_display *display, char *argv[], pid_t *pid_out, struct wl_event_source **sigchld_source)
+spawn_primary_client(struct cg_server *server, char *argv[], pid_t *pid_out, struct wl_event_source **sigchld_source)
 {
 	int fd[2];
 	if (pipe(fd) != 0) {
@@ -131,15 +152,15 @@ spawn_primary_client(struct wl_display *display, char *argv[], pid_t *pid_out, s
 	/* Close write, we only need read in Cage. */
 	close(fd[1]);
 
-	struct wl_event_loop *event_loop = wl_display_get_event_loop(display);
+	struct wl_event_loop *event_loop = wl_display_get_event_loop(server->wl_display);
 	uint32_t mask = WL_EVENT_HANGUP | WL_EVENT_ERROR;
-	*sigchld_source = wl_event_loop_add_fd(event_loop, fd[0], mask, sigchld_handler, display);
+	*sigchld_source = wl_event_loop_add_fd(event_loop, fd[0], mask, sigchld_handler, server);
 
 	wlr_log(WLR_DEBUG, "Child process created with pid %d", pid);
 	return true;
 }
 
-static void
+static int
 cleanup_primary_client(pid_t pid)
 {
 	int status;
@@ -148,9 +169,14 @@ cleanup_primary_client(pid_t pid)
 
 	if (WIFEXITED(status)) {
 		wlr_log(WLR_DEBUG, "Child exited normally with exit status %d", WEXITSTATUS(status));
+		return WEXITSTATUS(status);
 	} else if (WIFSIGNALED(status)) {
+		/* Mimic Bash and other shells for the exit status */
 		wlr_log(WLR_DEBUG, "Child was terminated by a signal (%d)", WTERMSIG(status));
+		return 128 + WTERMSIG(status);
 	}
+
+	return 0;
 }
 
 static bool
@@ -181,13 +207,13 @@ drop_permissions(void)
 static int
 handle_signal(int signal, void *data)
 {
-	struct wl_display *display = data;
+	struct cg_server *server = data;
 
 	switch (signal) {
 	case SIGINT:
 		/* Fallthrough */
 	case SIGTERM:
-		wl_display_terminate(display);
+		server_terminate(server);
 		return 0;
 	default:
 		return 0;
@@ -254,31 +280,9 @@ int
 main(int argc, char *argv[])
 {
 	struct cg_server server = {0};
-	struct wl_event_loop *event_loop = NULL;
-	struct wl_event_source *sigint_source = NULL;
-	struct wl_event_source *sigterm_source = NULL;
 	struct wl_event_source *sigchld_source = NULL;
-	struct wlr_compositor *compositor = NULL;
-	struct wlr_subcompositor *subcompositor = NULL;
-	struct wlr_data_device_manager *data_device_manager = NULL;
-	struct wlr_server_decoration_manager *server_decoration_manager = NULL;
-	struct wlr_xdg_decoration_manager_v1 *xdg_decoration_manager = NULL;
-	struct wlr_export_dmabuf_manager_v1 *export_dmabuf_manager = NULL;
-	struct wlr_screencopy_manager_v1 *screencopy_manager = NULL;
-	struct wlr_single_pixel_buffer_manager_v1 *single_pixel_buffer = NULL;
-	struct wlr_xdg_output_manager_v1 *output_manager = NULL;
-	struct wlr_gamma_control_manager_v1 *gamma_control_manager = NULL;
-	struct wlr_virtual_keyboard_manager_v1 *virtual_keyboard = NULL;
-	struct wlr_virtual_pointer_manager_v1 *virtual_pointer = NULL;
-	struct wlr_viewporter *viewporter = NULL;
-	struct wlr_presentation *presentation = NULL;
-	struct wlr_xdg_shell *xdg_shell = NULL;
-#if CAGE_HAS_XWAYLAND
-	struct wlr_xwayland *xwayland = NULL;
-	struct wlr_xcursor_manager *xcursor_manager = NULL;
-#endif
 	pid_t pid = 0;
-	int ret = 0;
+	int ret = 0, app_ret = 0;
 
 	if (!parse_args(&server, argc, argv)) {
 		return 1;
@@ -302,11 +306,14 @@ main(int argc, char *argv[])
 		return 1;
 	}
 
-	event_loop = wl_display_get_event_loop(server.wl_display);
-	sigint_source = wl_event_loop_add_signal(event_loop, SIGINT, handle_signal, &server.wl_display);
-	sigterm_source = wl_event_loop_add_signal(event_loop, SIGTERM, handle_signal, &server.wl_display);
+	server.display_destroy.notify = handle_display_destroy;
+	wl_display_add_destroy_listener(server.wl_display, &server.display_destroy);
 
-	server.backend = wlr_backend_autocreate(server.wl_display);
+	struct wl_event_loop *event_loop = wl_display_get_event_loop(server.wl_display);
+	struct wl_event_source *sigint_source = wl_event_loop_add_signal(event_loop, SIGINT, handle_signal, &server);
+	struct wl_event_source *sigterm_source = wl_event_loop_add_signal(event_loop, SIGTERM, handle_signal, &server);
+
+	server.backend = wlr_backend_autocreate(event_loop, &server.session);
 	if (!server.backend) {
 		wlr_log(WLR_ERROR, "Unable to create the wlroots backend");
 		ret = 1;
@@ -337,7 +344,7 @@ main(int argc, char *argv[])
 	wl_list_init(&server.views);
 	wl_list_init(&server.outputs);
 
-	server.output_layout = wlr_output_layout_create();
+	server.output_layout = wlr_output_layout_create(server.wl_display);
 	if (!server.output_layout) {
 		wlr_log(WLR_ERROR, "Unable to create output layout");
 		ret = 1;
@@ -353,29 +360,33 @@ main(int argc, char *argv[])
 		goto end;
 	}
 
-	wlr_scene_attach_output_layout(server.scene, server.output_layout);
+	server.scene_output_layout = wlr_scene_attach_output_layout(server.scene, server.output_layout);
 
-	compositor = wlr_compositor_create(server.wl_display, server.renderer);
+	struct wlr_compositor *compositor = wlr_compositor_create(server.wl_display, 6, server.renderer);
 	if (!compositor) {
 		wlr_log(WLR_ERROR, "Unable to create the wlroots compositor");
 		ret = 1;
 		goto end;
 	}
 
-	subcompositor = wlr_subcompositor_create(server.wl_display);
-	if (!subcompositor) {
+	if (!wlr_subcompositor_create(server.wl_display)) {
 		wlr_log(WLR_ERROR, "Unable to create the wlroots subcompositor");
 		ret = 1;
 		goto end;
 	}
 
-	data_device_manager = wlr_data_device_manager_create(server.wl_display);
-	if (!data_device_manager) {
+	if (!wlr_data_device_manager_create(server.wl_display)) {
 		wlr_log(WLR_ERROR, "Unable to create the data device manager");
 		ret = 1;
 		goto end;
 	}
 
+	if (!wlr_primary_selection_v1_device_manager_create(server.wl_display)) {
+		wlr_log(WLR_ERROR, "Unable to create primary selection device manager");
+		ret = 1;
+		goto end;
+	}
+
 	/* Configure a listener to be notified when new outputs are
 	 * available on the backend. We use this only to detect the
 	 * first output and ignore subsequent outputs. */
@@ -389,7 +400,7 @@ main(int argc, char *argv[])
 		goto end;
 	}
 
-	server.idle = wlr_idle_create(server.wl_display);
+	server.idle = wlr_idle_notifier_v1_create(server.wl_display);
 	if (!server.idle) {
 		wlr_log(WLR_ERROR, "Unable to create the idle tracker");
 		ret = 1;
@@ -406,16 +417,19 @@ main(int argc, char *argv[])
 	wl_signal_add(&server.idle_inhibit_v1->events.new_inhibitor, &server.new_idle_inhibitor_v1);
 	wl_list_init(&server.inhibitors);
 
-	xdg_shell = wlr_xdg_shell_create(server.wl_display, 4);
+	struct wlr_xdg_shell *xdg_shell = wlr_xdg_shell_create(server.wl_display, 4);
 	if (!xdg_shell) {
 		wlr_log(WLR_ERROR, "Unable to create the XDG shell interface");
 		ret = 1;
 		goto end;
 	}
-	server.new_xdg_shell_surface.notify = handle_xdg_shell_surface_new;
-	wl_signal_add(&xdg_shell->events.new_surface, &server.new_xdg_shell_surface);
+	server.new_xdg_toplevel.notify = handle_new_xdg_toplevel;
+	wl_signal_add(&xdg_shell->events.new_toplevel, &server.new_xdg_toplevel);
+	server.new_xdg_popup.notify = handle_new_xdg_popup;
+	wl_signal_add(&xdg_shell->events.new_popup, &server.new_xdg_popup);
 
-	xdg_decoration_manager = wlr_xdg_decoration_manager_v1_create(server.wl_display);
+	struct wlr_xdg_decoration_manager_v1 *xdg_decoration_manager =
+		wlr_xdg_decoration_manager_v1_create(server.wl_display);
 	if (!xdg_decoration_manager) {
 		wlr_log(WLR_ERROR, "Unable to create the XDG decoration manager");
 		ret = 1;
@@ -424,7 +438,8 @@ main(int argc, char *argv[])
 	wl_signal_add(&xdg_decoration_manager->events.new_toplevel_decoration, &server.xdg_toplevel_decoration);
 	server.xdg_toplevel_decoration.notify = handle_xdg_toplevel_decoration;
 
-	server_decoration_manager = wlr_server_decoration_manager_create(server.wl_display);
+	struct wlr_server_decoration_manager *server_decoration_manager =
+		wlr_server_decoration_manager_create(server.wl_display);
 	if (!server_decoration_manager) {
 		wlr_log(WLR_ERROR, "Unable to create the server decoration manager");
 		ret = 1;
@@ -434,44 +449,38 @@ main(int argc, char *argv[])
 		server_decoration_manager, server.xdg_decoration ? WLR_SERVER_DECORATION_MANAGER_MODE_SERVER
 								 : WLR_SERVER_DECORATION_MANAGER_MODE_CLIENT);
 
-	viewporter = wlr_viewporter_create(server.wl_display);
-	if (!viewporter) {
+	if (!wlr_viewporter_create(server.wl_display)) {
 		wlr_log(WLR_ERROR, "Unable to create the viewporter interface");
 		ret = 1;
 		goto end;
 	}
 
-	presentation = wlr_presentation_create(server.wl_display, server.backend);
+	struct wlr_presentation *presentation = wlr_presentation_create(server.wl_display, server.backend);
 	if (!presentation) {
 		wlr_log(WLR_ERROR, "Unable to create the presentation interface");
 		ret = 1;
 		goto end;
 	}
-	wlr_scene_set_presentation(server.scene, presentation);
 
-	export_dmabuf_manager = wlr_export_dmabuf_manager_v1_create(server.wl_display);
-	if (!export_dmabuf_manager) {
+	if (!wlr_export_dmabuf_manager_v1_create(server.wl_display)) {
 		wlr_log(WLR_ERROR, "Unable to create the export DMABUF manager");
 		ret = 1;
 		goto end;
 	}
 
-	screencopy_manager = wlr_screencopy_manager_v1_create(server.wl_display);
-	if (!screencopy_manager) {
+	if (!wlr_screencopy_manager_v1_create(server.wl_display)) {
 		wlr_log(WLR_ERROR, "Unable to create the screencopy manager");
 		ret = 1;
 		goto end;
 	}
 
-	single_pixel_buffer = wlr_single_pixel_buffer_manager_v1_create(server.wl_display);
-	if (!single_pixel_buffer) {
+	if (!wlr_single_pixel_buffer_manager_v1_create(server.wl_display)) {
 		wlr_log(WLR_ERROR, "Unable to create the single pixel buffer manager");
 		ret = 1;
 		goto end;
 	}
 
-	output_manager = wlr_xdg_output_manager_v1_create(server.wl_display, server.output_layout);
-	if (!output_manager) {
+	if (!wlr_xdg_output_manager_v1_create(server.wl_display, server.output_layout)) {
 		wlr_log(WLR_ERROR, "Unable to create the output manager");
 		ret = 1;
 		goto end;
@@ -488,14 +497,14 @@ main(int argc, char *argv[])
 	server.output_manager_test.notify = handle_output_manager_test;
 	wl_signal_add(&server.output_manager_v1->events.test, &server.output_manager_test);
 
-	gamma_control_manager = wlr_gamma_control_manager_v1_create(server.wl_display);
-	if (!gamma_control_manager) {
+	if (!wlr_gamma_control_manager_v1_create(server.wl_display)) {
 		wlr_log(WLR_ERROR, "Unable to create the gamma control manager");
 		ret = 1;
 		goto end;
 	}
 
-	virtual_keyboard = wlr_virtual_keyboard_manager_v1_create(server.wl_display);
+	struct wlr_virtual_keyboard_manager_v1 *virtual_keyboard =
+		wlr_virtual_keyboard_manager_v1_create(server.wl_display);
 	if (!virtual_keyboard) {
 		wlr_log(WLR_ERROR, "Unable to create the virtual keyboard manager");
 		ret = 1;
@@ -503,7 +512,8 @@ main(int argc, char *argv[])
 	}
 	wl_signal_add(&virtual_keyboard->events.new_virtual_keyboard, &server.new_virtual_keyboard);
 
-	virtual_pointer = wlr_virtual_pointer_manager_v1_create(server.wl_display);
+	struct wlr_virtual_pointer_manager_v1 *virtual_pointer =
+		wlr_virtual_pointer_manager_v1_create(server.wl_display);
 	if (!virtual_pointer) {
 		wlr_log(WLR_ERROR, "Unable to create the virtual pointer manager");
 		ret = 1;
@@ -511,37 +521,45 @@ main(int argc, char *argv[])
 	}
 	wl_signal_add(&virtual_pointer->events.new_virtual_pointer, &server.new_virtual_pointer);
 
+	server.relative_pointer_manager = wlr_relative_pointer_manager_v1_create(server.wl_display);
+	if (!server.relative_pointer_manager) {
+		wlr_log(WLR_ERROR, "Unable to create the relative pointer manager");
+		ret = 1;
+		goto end;
+	}
+
 #if CAGE_HAS_XWAYLAND
-	xwayland = wlr_xwayland_create(server.wl_display, compositor, true);
+	struct wlr_xcursor_manager *xcursor_manager = NULL;
+	struct wlr_xwayland *xwayland = wlr_xwayland_create(server.wl_display, compositor, true);
 	if (!xwayland) {
 		wlr_log(WLR_ERROR, "Cannot create XWayland server");
-		ret = 1;
-		goto end;
-	}
-	server.new_xwayland_surface.notify = handle_xwayland_surface_new;
-	wl_signal_add(&xwayland->events.new_surface, &server.new_xwayland_surface);
-
-	xcursor_manager = wlr_xcursor_manager_create(DEFAULT_XCURSOR, XCURSOR_SIZE);
-	if (!xcursor_manager) {
-		wlr_log(WLR_ERROR, "Cannot create XWayland XCursor manager");
-		ret = 1;
-		goto end;
-	}
-
-	if (setenv("DISPLAY", xwayland->display_name, true) < 0) {
-		wlr_log_errno(WLR_ERROR, "Unable to set DISPLAY for XWayland. Clients may not be able to connect");
 	} else {
-		wlr_log(WLR_DEBUG, "XWayland is running on display %s", xwayland->display_name);
-	}
+		server.new_xwayland_surface.notify = handle_xwayland_surface_new;
+		wl_signal_add(&xwayland->events.new_surface, &server.new_xwayland_surface);
 
-	if (!wlr_xcursor_manager_load(xcursor_manager, 1)) {
-		wlr_log(WLR_ERROR, "Cannot load XWayland XCursor theme");
-	}
-	struct wlr_xcursor *xcursor = wlr_xcursor_manager_get_xcursor(xcursor_manager, DEFAULT_XCURSOR, 1);
-	if (xcursor) {
-		struct wlr_xcursor_image *image = xcursor->images[0];
-		wlr_xwayland_set_cursor(xwayland, image->buffer, image->width * 4, image->width, image->height,
-					image->hotspot_x, image->hotspot_y);
+		xcursor_manager = wlr_xcursor_manager_create(DEFAULT_XCURSOR, XCURSOR_SIZE);
+		if (!xcursor_manager) {
+			wlr_log(WLR_ERROR, "Cannot create XWayland XCursor manager");
+			ret = 1;
+			goto end;
+		}
+
+		if (setenv("DISPLAY", xwayland->display_name, true) < 0) {
+			wlr_log_errno(WLR_ERROR,
+				      "Unable to set DISPLAY for XWayland. Clients may not be able to connect");
+		} else {
+			wlr_log(WLR_DEBUG, "XWayland is running on display %s", xwayland->display_name);
+		}
+
+		if (!wlr_xcursor_manager_load(xcursor_manager, 1)) {
+			wlr_log(WLR_ERROR, "Cannot load XWayland XCursor theme");
+		}
+		struct wlr_xcursor *xcursor = wlr_xcursor_manager_get_xcursor(xcursor_manager, DEFAULT_XCURSOR, 1);
+		if (xcursor) {
+			struct wlr_xcursor_image *image = xcursor->images[0];
+			wlr_xwayland_set_cursor(xwayland, image->buffer, image->width * 4, image->width, image->height,
+						image->hotspot_x, image->hotspot_y);
+		}
 	}
 #endif
 
@@ -565,19 +583,17 @@ main(int argc, char *argv[])
 	}
 
 #if CAGE_HAS_XWAYLAND
-	wlr_xwayland_set_seat(xwayland, server.seat->seat);
+	if (xwayland) {
+		wlr_xwayland_set_seat(xwayland, server.seat->seat);
+	}
 #endif
 
-	if (!spawn_primary_client(server.wl_display, argv + optind, &pid, &sigchld_source)) {
+	if (!spawn_primary_client(&server, argv + optind, &pid, &sigchld_source)) {
 		ret = 1;
 		goto end;
 	}
 
-	/* Place the cursor in the center of the output layout. */
-	struct wlr_box layout_box;
-	wlr_output_layout_get_box(server.output_layout, NULL, &layout_box);
-	wlr_cursor_warp(server.seat->cursor, NULL, layout_box.width / 2, layout_box.height / 2);
-
+	seat_center_cursor(server.seat);
 	wl_display_run(server.wl_display);
 
 #if CAGE_HAS_XWAYLAND
@@ -587,7 +603,9 @@ main(int argc, char *argv[])
 	wl_display_destroy_clients(server.wl_display);
 
 end:
-	cleanup_primary_client(pid);
+	app_ret = cleanup_primary_client(pid);
+	if (!ret && server.return_app_code)
+		ret = app_ret;
 
 	wl_event_source_remove(sigint_source);
 	wl_event_source_remove(sigterm_source);
@@ -598,6 +616,5 @@ end:
 	/* This function is not null-safe, but we only ever get here
 	   with a proper wl_display. */
 	wl_display_destroy(server.wl_display);
-	wlr_output_layout_destroy(server.output_layout);
 	return ret;
 }
