@@ -26,7 +26,6 @@
 #include <wlr/types/wlr_data_device.h>
 #include <wlr/types/wlr_matrix.h>
 #include <wlr/types/wlr_output.h>
-#include <wlr/types/wlr_output_damage.h>
 #include <wlr/types/wlr_output_layout.h>
 #include <wlr/types/wlr_output_management_v1.h>
 #include <wlr/types/wlr_scene.h>
@@ -43,7 +42,7 @@
 #endif
 
 #define OUTPUT_CONFIG_UPDATED                                                                                          \
-	(WLR_OUTPUT_STATE_ENABLED | WLR_OUTPUT_STATE_SCALE | WLR_OUTPUT_STATE_TRANSFORM |                              \
+	(WLR_OUTPUT_STATE_ENABLED | WLR_OUTPUT_STATE_MODE | WLR_OUTPUT_STATE_SCALE | WLR_OUTPUT_STATE_TRANSFORM |      \
 	 WLR_OUTPUT_STATE_ADAPTIVE_SYNC_ENABLED)
 
 static void
@@ -68,6 +67,34 @@ update_output_manager_config(struct cg_server *server)
 	wlr_output_manager_v1_set_configuration(server->output_manager_v1, config);
 }
 
+static inline void
+output_layout_add_auto(struct cg_output *output)
+{
+	assert(output->scene_output != NULL);
+	struct wlr_output_layout_output *layout_output =
+		wlr_output_layout_add_auto(output->server->output_layout, output->wlr_output);
+	wlr_scene_output_layout_add_output(output->server->scene_output_layout, layout_output, output->scene_output);
+}
+
+static inline void
+output_layout_add(struct cg_output *output, int32_t x, int32_t y)
+{
+	assert(output->scene_output != NULL);
+	bool exists = wlr_output_layout_get(output->server->output_layout, output->wlr_output);
+	struct wlr_output_layout_output *layout_output =
+		wlr_output_layout_add(output->server->output_layout, output->wlr_output, x, y);
+	if (exists) {
+		return;
+	}
+	wlr_scene_output_layout_add_output(output->server->scene_output_layout, layout_output, output->scene_output);
+}
+
+static inline void
+output_layout_remove(struct cg_output *output)
+{
+	wlr_output_layout_remove(output->server->output_layout, output->wlr_output);
+}
+
 static void
 output_enable(struct cg_output *output)
 {
@@ -78,12 +105,12 @@ output_enable(struct cg_output *output)
 	 * duplicate the enabled property in cg_output. */
 	wlr_log(WLR_DEBUG, "Enabling output %s", wlr_output->name);
 
-	wlr_output_layout_add_auto(output->server->output_layout, wlr_output);
-	wlr_output_enable(wlr_output, true);
-	wlr_output_commit(wlr_output);
+	struct wlr_output_state state = {0};
+	wlr_output_state_set_enabled(&state, true);
 
-	output->scene_output = wlr_scene_get_scene_output(output->server->scene, wlr_output);
-	assert(output->scene_output != NULL);
+	if (wlr_output_commit_state(wlr_output, &state)) {
+		output_layout_add_auto(output);
+	}
 
 	update_output_manager_config(output->server);
 }
@@ -92,53 +119,38 @@ static void
 output_disable(struct cg_output *output)
 {
 	struct wlr_output *wlr_output = output->wlr_output;
-
 	if (!wlr_output->enabled) {
 		wlr_log(WLR_DEBUG, "Not disabling already disabled output %s", wlr_output->name);
 		return;
 	}
 
-	output->scene_output = NULL;
-
 	wlr_log(WLR_DEBUG, "Disabling output %s", wlr_output->name);
-	wlr_output_enable(wlr_output, false);
-	wlr_output_layout_remove(output->server->output_layout, wlr_output);
-	wlr_output_commit(wlr_output);
+	struct wlr_output_state state = {0};
+	wlr_output_state_set_enabled(&state, false);
+	wlr_output_commit_state(wlr_output, &state);
+	output_layout_remove(output);
 }
 
 static bool
 output_apply_config(struct cg_output *output, struct wlr_output_configuration_head_v1 *head, bool test_only)
 {
-	wlr_output_enable(output->wlr_output, head->state.enabled);
-
-	if (head->state.enabled) {
-		/* Do not mess with these parameters for output to be disabled */
-		wlr_output_set_scale(output->wlr_output, head->state.scale);
-		wlr_output_set_transform(output->wlr_output, head->state.transform);
-
-		if (head->state.mode) {
-			wlr_output_set_mode(output->wlr_output, head->state.mode);
-		} else {
-			wlr_output_set_custom_mode(output->wlr_output, head->state.custom_mode.width,
-						   head->state.custom_mode.height, head->state.custom_mode.refresh);
-		}
-	}
+	struct wlr_output_state state = {0};
+	wlr_output_head_v1_state_apply(&head->state, &state);
 
 	if (test_only) {
-		bool ret = wlr_output_test(output->wlr_output);
-		wlr_output_rollback(output->wlr_output);
+		bool ret = wlr_output_test_state(output->wlr_output, &state);
 		return ret;
 	}
 
 	/* Apply output configuration */
-	if (!wlr_output_commit(output->wlr_output)) {
+	if (!wlr_output_commit_state(output->wlr_output, &state)) {
 		return false;
 	}
 
 	if (head->state.enabled) {
-		wlr_output_layout_add(output->server->output_layout, head->state.output, head->state.x, head->state.y);
+		output_layout_add(output, head->state.x, head->state.y);
 	} else {
-		wlr_output_layout_remove(output->server->output_layout, output->wlr_output);
+		output_layout_remove(output);
 	}
 
 	return true;
@@ -149,11 +161,11 @@ handle_output_frame(struct wl_listener *listener, void *data)
 {
 	struct cg_output *output = wl_container_of(listener, output, frame);
 
-	if (!output->wlr_output->enabled) {
+	if (!output->wlr_output->enabled || !output->scene_output) {
 		return;
 	}
 
-	wlr_scene_output_commit(output->scene_output);
+	wlr_scene_output_commit(output->scene_output, NULL);
 
 	struct timespec now = {0};
 	clock_gettime(CLOCK_MONOTONIC, &now);
@@ -170,31 +182,20 @@ handle_output_commit(struct wl_listener *listener, void *data)
 	 * - output layout change will also be called if needed to position the views
 	 * - always update output manager configuration even if the output is now disabled */
 
-	if (event->committed & WLR_OUTPUT_STATE_ENABLED) {
-		if (output->wlr_output->enabled) {
-			output->scene_output = wlr_scene_get_scene_output(output->server->scene, output->wlr_output);
-			assert(output->scene_output != NULL);
-		} else {
-			output->scene_output = NULL;
-		}
-	}
-
-	if (event->committed & OUTPUT_CONFIG_UPDATED) {
+	if (event->state->committed & OUTPUT_CONFIG_UPDATED) {
 		update_output_manager_config(output->server);
 	}
 }
 
 static void
-handle_output_mode(struct wl_listener *listener, void *data)
+handle_output_request_state(struct wl_listener *listener, void *data)
 {
-	struct cg_output *output = wl_container_of(listener, output, mode);
+	struct cg_output *output = wl_container_of(listener, output, request_state);
+	struct wlr_output_event_request_state *event = data;
 
-	if (!output->wlr_output->enabled) {
-		return;
+	if (wlr_output_commit_state(output->wlr_output, event->state)) {
+		update_output_manager_config(output->server);
 	}
-
-	view_position_all(output->server);
-	update_output_manager_config(output->server);
 }
 
 void
@@ -206,25 +207,40 @@ handle_output_layout_change(struct wl_listener *listener, void *data)
 	update_output_manager_config(server);
 }
 
+static bool
+is_nested_output(struct cg_output *output)
+{
+	if (wlr_output_is_wl(output->wlr_output)) {
+		return true;
+	}
+#if WLR_HAS_X11_BACKEND
+	if (wlr_output_is_x11(output->wlr_output)) {
+		return true;
+	}
+#endif
+	return false;
+}
+
 static void
 output_destroy(struct cg_output *output)
 {
 	struct cg_server *server = output->server;
+	bool was_nested_output = is_nested_output(output);
 
 	output->wlr_output->data = NULL;
 
 	wl_list_remove(&output->destroy.link);
 	wl_list_remove(&output->commit.link);
-	wl_list_remove(&output->mode.link);
+	wl_list_remove(&output->request_state.link);
 	wl_list_remove(&output->frame.link);
 	wl_list_remove(&output->link);
 
-	wlr_output_layout_remove(server->output_layout, output->wlr_output);
+	output_layout_remove(output);
 
 	free(output);
 
-	if (wl_list_empty(&server->outputs)) {
-		wl_display_terminate(server->wl_display);
+	if (wl_list_empty(&server->outputs) && was_nested_output) {
+		server_terminate(server);
 	} else if (server->output_mode == CAGE_MULTI_OUTPUT_MODE_LAST && !wl_list_empty(&server->outputs)) {
 		struct cg_output *prev = wl_container_of(server->outputs.next, prev, link);
 		output_enable(prev);
@@ -264,27 +280,35 @@ handle_new_output(struct wl_listener *listener, void *data)
 
 	output->commit.notify = handle_output_commit;
 	wl_signal_add(&wlr_output->events.commit, &output->commit);
-	output->mode.notify = handle_output_mode;
-	wl_signal_add(&wlr_output->events.mode, &output->mode);
+	output->request_state.notify = handle_output_request_state;
+	wl_signal_add(&wlr_output->events.request_state, &output->request_state);
 	output->destroy.notify = handle_output_destroy;
 	wl_signal_add(&wlr_output->events.destroy, &output->destroy);
 	output->frame.notify = handle_output_frame;
 	wl_signal_add(&wlr_output->events.frame, &output->frame);
 
+	output->scene_output = wlr_scene_output_create(server->scene, wlr_output);
+	if (!output->scene_output) {
+		wlr_log(WLR_ERROR, "Failed to allocate scene output");
+		return;
+	}
+
+	struct wlr_output_state state = {0};
+	wlr_output_state_set_enabled(&state, true);
 	if (!wl_list_empty(&wlr_output->modes)) {
 		struct wlr_output_mode *preferred_mode = wlr_output_preferred_mode(wlr_output);
 		if (preferred_mode) {
-			wlr_output_set_mode(wlr_output, preferred_mode);
+			wlr_output_state_set_mode(&state, preferred_mode);
 		}
-		if (!wlr_output_test(wlr_output)) {
+		if (!wlr_output_test_state(wlr_output, &state)) {
 			struct wlr_output_mode *mode;
 			wl_list_for_each (mode, &wlr_output->modes, link) {
 				if (mode == preferred_mode) {
 					continue;
 				}
 
-				wlr_output_set_mode(wlr_output, mode);
-				if (wlr_output_test(wlr_output)) {
+				wlr_output_state_set_mode(&state, mode);
+				if (wlr_output_test_state(wlr_output, &state)) {
 					break;
 				}
 			}
@@ -301,8 +325,13 @@ handle_new_output(struct wl_listener *listener, void *data)
 			wlr_output->scale);
 	}
 
-	output_enable(output);
+	wlr_log(WLR_DEBUG, "Enabling new output %s", wlr_output->name);
+	if (wlr_output_commit_state(wlr_output, &state)) {
+		output_layout_add_auto(output);
+	}
+
 	view_position_all(output->server);
+	update_output_manager_config(output->server);
 }
 
 void
