@@ -17,11 +17,13 @@
 #include <wayland-server-core.h>
 #include <wlr/backend.h>
 #include <wlr/backend/multi.h>
+#include <wlr/backend/session.h>
 #include <wlr/types/wlr_cursor.h>
 #include <wlr/types/wlr_data_device.h>
-#include <wlr/types/wlr_idle.h>
+#include <wlr/types/wlr_idle_notify_v1.h>
 #include <wlr/types/wlr_keyboard_group.h>
 #include <wlr/types/wlr_primary_selection.h>
+#include <wlr/types/wlr_relative_pointer_v1.h>
 #include <wlr/types/wlr_scene.h>
 #include <wlr/types/wlr_seat.h>
 #include <wlr/types/wlr_touch.h>
@@ -63,7 +65,7 @@ desktop_view_at(struct cg_server *server, double lx, double ly, struct wlr_surfa
 	}
 
 	struct wlr_scene_buffer *scene_buffer = wlr_scene_buffer_from_node(node);
-	struct wlr_scene_surface *scene_surface = wlr_scene_surface_from_buffer(scene_buffer);
+	struct wlr_scene_surface *scene_surface = wlr_scene_surface_try_from_buffer(scene_buffer);
 	if (!scene_surface) {
 		return NULL;
 	}
@@ -126,9 +128,9 @@ update_capabilities(struct cg_seat *seat)
 
 	/* Hide cursor if the seat doesn't have pointer capability. */
 	if ((caps & WL_SEAT_CAPABILITY_POINTER) == 0) {
-		wlr_cursor_set_image(seat->cursor, NULL, 0, 0, 0, 0, 0, 0);
+		wlr_cursor_unset_image(seat->cursor);
 	} else {
-		wlr_xcursor_manager_set_cursor_image(seat->xcursor_manager, DEFAULT_XCURSOR, seat->cursor);
+		wlr_cursor_set_xcursor(seat->cursor, seat->xcursor_manager, DEFAULT_XCURSOR);
 	}
 }
 
@@ -248,7 +250,7 @@ handle_modifier_event(struct wlr_keyboard *keyboard, struct cg_seat *seat)
 	wlr_seat_set_keyboard(seat->seat, keyboard);
 	wlr_seat_keyboard_notify_modifiers(seat->seat, &keyboard->modifiers);
 
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static bool
@@ -256,22 +258,21 @@ handle_keybinding(struct cg_server *server, xkb_keysym_t sym)
 {
 #ifdef DEBUG
 	if (sym == XKB_KEY_Escape) {
-		wl_display_terminate(server->wl_display);
+		server_terminate(server);
 		return true;
 	}
 #endif
 	if (server->allow_vt_switch && sym >= XKB_KEY_XF86Switch_VT_1 && sym <= XKB_KEY_XF86Switch_VT_12) {
 		if (wlr_backend_is_multi(server->backend)) {
-			struct wlr_session *session = wlr_backend_get_session(server->backend);
-			if (session) {
+			if (server->session) {
 				unsigned vt = sym - XKB_KEY_XF86Switch_VT_1 + 1;
-				wlr_session_change_vt(session, vt);
+				wlr_session_change_vt(server->session, vt);
 			}
 		}
 	} else {
 		return false;
 	}
-	wlr_idle_notify_activity(server->idle, server->seat->seat);
+	wlr_idle_notifier_v1_notify_activity(server->idle, server->seat->seat);
 	return true;
 }
 
@@ -303,7 +304,7 @@ handle_key_event(struct wlr_keyboard *keyboard, struct cg_seat *seat, void *data
 		wlr_seat_keyboard_notify_key(seat->seat, event->time_msec, event->keycode, event->state);
 	}
 
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
@@ -440,7 +441,7 @@ handle_new_input(struct wl_listener *listener, void *data)
 	case WLR_INPUT_DEVICE_SWITCH:
 		wlr_log(WLR_DEBUG, "Switch input is not implemented");
 		return;
-	case WLR_INPUT_DEVICE_TABLET_TOOL:
+	case WLR_INPUT_DEVICE_TABLET:
 	case WLR_INPUT_DEVICE_TABLET_PAD:
 		wlr_log(WLR_DEBUG, "Tablet input is not implemented");
 		return;
@@ -511,7 +512,7 @@ handle_touch_down(struct wl_listener *listener, void *data)
 		press_cursor_button(seat, &event->touch->base, event->time_msec, BTN_LEFT, WLR_BUTTON_PRESSED, lx, ly);
 	}
 
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
@@ -530,7 +531,7 @@ handle_touch_up(struct wl_listener *listener, void *data)
 	}
 
 	wlr_seat_touch_notify_up(seat->seat, event->time_msec, event->touch_id);
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
@@ -562,7 +563,16 @@ handle_touch_motion(struct wl_listener *listener, void *data)
 		seat->touch_ly = ly;
 	}
 
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
+}
+
+static void
+handle_touch_frame(struct wl_listener *listener, void *data)
+{
+	struct cg_seat *seat = wl_container_of(listener, seat, touch_frame);
+
+	wlr_seat_touch_notify_frame(seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
@@ -571,7 +581,7 @@ handle_cursor_frame(struct wl_listener *listener, void *data)
 	struct cg_seat *seat = wl_container_of(listener, seat, cursor_frame);
 
 	wlr_seat_pointer_notify_frame(seat->seat);
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
@@ -581,8 +591,8 @@ handle_cursor_axis(struct wl_listener *listener, void *data)
 	struct wlr_pointer_axis_event *event = data;
 
 	wlr_seat_pointer_notify_axis(seat->seat, event->time_msec, event->orientation, event->delta,
-				     event->delta_discrete, event->source);
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+				     event->delta_discrete, event->source, event->relative_direction);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
@@ -594,24 +604,29 @@ handle_cursor_button(struct wl_listener *listener, void *data)
 	wlr_seat_pointer_notify_button(seat->seat, event->time_msec, event->button, event->state);
 	press_cursor_button(seat, &event->pointer->base, event->time_msec, event->button, event->state, seat->cursor->x,
 			    seat->cursor->y);
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
-process_cursor_motion(struct cg_seat *seat, uint32_t time)
+process_cursor_motion(struct cg_seat *seat, uint32_t time_msec, double dx, double dy, double dx_unaccel,
+		      double dy_unaccel)
 {
 	double sx, sy;
 	struct wlr_seat *wlr_seat = seat->seat;
 	struct wlr_surface *surface = NULL;
 
 	struct cg_view *view = desktop_view_at(seat->server, seat->cursor->x, seat->cursor->y, &surface, &sx, &sy);
-
 	if (!view) {
 		wlr_seat_pointer_clear_focus(wlr_seat);
 	} else {
 		wlr_seat_pointer_notify_enter(wlr_seat, surface, sx, sy);
+		wlr_seat_pointer_notify_motion(wlr_seat, time_msec, sx, sy);
+	}
 
-		wlr_seat_pointer_notify_motion(wlr_seat, time, sx, sy);
+	if (dx != 0 || dy != 0) {
+		wlr_relative_pointer_manager_v1_send_relative_motion(seat->server->relative_pointer_manager, wlr_seat,
+								     (uint64_t) time_msec * 1000, dx, dy, dx_unaccel,
+								     dy_unaccel);
 	}
 
 	struct cg_drag_icon *drag_icon;
@@ -619,7 +634,7 @@ process_cursor_motion(struct cg_seat *seat, uint32_t time)
 		drag_icon_update_position(drag_icon);
 	}
 
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
@@ -628,20 +643,27 @@ handle_cursor_motion_absolute(struct wl_listener *listener, void *data)
 	struct cg_seat *seat = wl_container_of(listener, seat, cursor_motion_absolute);
 	struct wlr_pointer_motion_absolute_event *event = data;
 
+	double lx, ly;
+	wlr_cursor_absolute_to_layout_coords(seat->cursor, &event->pointer->base, event->x, event->y, &lx, &ly);
+
+	double dx = lx - seat->cursor->x;
+	double dy = ly - seat->cursor->y;
+
 	wlr_cursor_warp_absolute(seat->cursor, &event->pointer->base, event->x, event->y);
-	process_cursor_motion(seat, event->time_msec);
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	process_cursor_motion(seat, event->time_msec, dx, dy, dx, dy);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
-handle_cursor_motion(struct wl_listener *listener, void *data)
+handle_cursor_motion_relative(struct wl_listener *listener, void *data)
 {
-	struct cg_seat *seat = wl_container_of(listener, seat, cursor_motion);
+	struct cg_seat *seat = wl_container_of(listener, seat, cursor_motion_relative);
 	struct wlr_pointer_motion_event *event = data;
 
 	wlr_cursor_move(seat->cursor, &event->pointer->base, event->delta_x, event->delta_y);
-	process_cursor_motion(seat, event->time_msec);
-	wlr_idle_notify_activity(seat->server->idle, seat->seat);
+	process_cursor_motion(seat, event->time_msec, event->delta_x, event->delta_y, event->unaccel_dx,
+			      event->unaccel_dy);
+	wlr_idle_notifier_v1_notify_activity(seat->server->idle, seat->seat);
 }
 
 static void
@@ -740,7 +762,7 @@ handle_destroy(struct wl_listener *listener, void *data)
 {
 	struct cg_seat *seat = wl_container_of(listener, seat, destroy);
 	wl_list_remove(&seat->destroy.link);
-	wl_list_remove(&seat->cursor_motion.link);
+	wl_list_remove(&seat->cursor_motion_relative.link);
 	wl_list_remove(&seat->cursor_motion_absolute.link);
 	wl_list_remove(&seat->cursor_button.link);
 	wl_list_remove(&seat->cursor_axis.link);
@@ -748,6 +770,7 @@ handle_destroy(struct wl_listener *listener, void *data)
 	wl_list_remove(&seat->touch_down.link);
 	wl_list_remove(&seat->touch_up.link);
 	wl_list_remove(&seat->touch_motion.link);
+	wl_list_remove(&seat->touch_frame.link);
 	wl_list_remove(&seat->request_set_cursor.link);
 	wl_list_remove(&seat->request_set_selection.link);
 	wl_list_remove(&seat->request_set_primary_selection.link);
@@ -813,8 +836,8 @@ seat_create(struct cg_server *server, struct wlr_backend *backend)
 		}
 	}
 
-	seat->cursor_motion.notify = handle_cursor_motion;
-	wl_signal_add(&seat->cursor->events.motion, &seat->cursor_motion);
+	seat->cursor_motion_relative.notify = handle_cursor_motion_relative;
+	wl_signal_add(&seat->cursor->events.motion, &seat->cursor_motion_relative);
 	seat->cursor_motion_absolute.notify = handle_cursor_motion_absolute;
 	wl_signal_add(&seat->cursor->events.motion_absolute, &seat->cursor_motion_absolute);
 	seat->cursor_button.notify = handle_cursor_button;
@@ -830,6 +853,8 @@ seat_create(struct cg_server *server, struct wlr_backend *backend)
 	wl_signal_add(&seat->cursor->events.touch_up, &seat->touch_up);
 	seat->touch_motion.notify = handle_touch_motion;
 	wl_signal_add(&seat->cursor->events.touch_motion, &seat->touch_motion);
+	seat->touch_frame.notify = handle_touch_frame;
+	wl_signal_add(&seat->cursor->events.touch_frame, &seat->touch_frame);
 
 	seat->request_set_cursor.notify = handle_request_set_cursor;
 	wl_signal_add(&seat->seat->events.request_set_cursor, &seat->request_set_cursor);
@@ -930,5 +955,14 @@ seat_set_focus(struct cg_seat *seat, struct cg_view *view)
 		wlr_seat_keyboard_notify_enter(wlr_seat, view->wlr_surface, NULL, 0, NULL);
 	}
 
-	process_cursor_motion(seat, -1);
+	process_cursor_motion(seat, -1, 0, 0, 0, 0);
+}
+
+void
+seat_center_cursor(struct cg_seat *seat)
+{
+	/* Place the cursor in the center of the output layout. */
+	struct wlr_box layout_box;
+	wlr_output_layout_get_box(seat->server->output_layout, NULL, &layout_box);
+	wlr_cursor_warp(seat->cursor, NULL, layout_box.width / 2, layout_box.height / 2);
 }
